SystemOrganization addCategory: #'Neo-UUID'!

TestCase subclass: #NeoUUIDGeneratorTests
	instanceVariableNames: 'generator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-UUID'!

!NeoUUIDGeneratorTests commentStamp: '<historical>' prior: 0!
NeoUUIDGeneratorTests contains units tests for NeoUUIDGenerator.
!

!NeoUUIDGeneratorTests methodsFor: 'setup' stamp: 'SvenVanCaekenberghe 8/31/2013 14:40'!
setUp
	generator := NeoUUIDGenerator new! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:37'!
testDefault
	self assert: NeoUUIDGenerator default equals: NeoUUIDGenerator default.
	self deny: NeoUUIDGenerator next isNilUUID.
	self assert: (NeoUUIDGenerator default randomGenerator next between: 0 and: 1)! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/7/2016 11:32'!
testNextRandom16
	| set |
	set := Set new.
	1e6 timesRepeat: [ set add: NeoUUIDGenerator default nextRandom16 ].
	self assert: set equals: (0 to: 16rFFFF) asSet! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/7/2016 11:31'!
testNextRandom8
	| set |
	set := Set new.
	1e4 timesRepeat: [ set add: NeoUUIDGenerator default nextRandom8 ].
	self assert: set equals: (0 to: 255) asSet! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/31/2013 14:41'!
testOne
	| uuid |
	uuid := generator next.
	self assert: (uuid isKindOf: UUID).
	self deny: uuid isNilUUID! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:38'!
testOneRandom
	| uuid |
	uuid := generator nextRandom.
	self assert: (uuid isKindOf: UUID).
	self deny: uuid isNilUUID! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:35'!
testRandomBytes
	| bytes |
	bytes := generator nextRandomByteArrayOfSize: 32.
	self assert: (bytes isKindOf: ByteArray).
	self assert: bytes size equals: 32! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/31/2013 14:46'!
testSpeed
	self 
		should: [ 100000 timesRepeat: [ generator next ] ] 
		notTakeMoreThan: 1 second! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2016 12:00'!
testTwoDifferentGenerator
	| generator1 generator2 uuid1 uuid2 time1 time2 |
	generator1 := NeoUUIDGenerator new.
	generator2 := NeoUUIDGenerator new.
	uuid1 := generator1 next.
	uuid2 := generator2 next.
	"different UUIDs"
	self deny: uuid1 = uuid2.
	"different genertors, different node id"
	self deny: uuid1 reverse readStream uint32 = uuid2 reverse readStream uint32.
	"different genertors, similar clock, within 100 microseconds (time is in milliseconds)"
	time1 := self timeFromUUID: uuid1.
	time2 := self timeFromUUID: uuid2.
	self assert: (time2 - time1) abs < (100 * 1000)! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/11/2016 11:59'!
testTwoSameGenerator
	| uuid1 uuid2 time1 time2 |
	uuid1 := generator next.
	uuid2 := generator next.
	"different UUIDs"
	self deny: uuid1 = uuid2.
	"same genertor, same node id"
	self assert: uuid1 reverse readStream uint32 equals: uuid2 reverse readStream uint32.
	"same generator, consequetive counters (take roll-over into account)"
	self assert: ((uuid1 at: 10) + 1) \\ 256 equals: (uuid2 at: 10).
	"same genertors, similar clock, within 100 microseconds (time is in milliseconds)"
	time1 := self timeFromUUID: uuid1.
	time2 := self timeFromUUID: uuid2.
	self assert: (time2 - time1) abs < (100 * 1000)! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 8/31/2013 14:44'!
testUniqueness
	| uuids count |
	count := 100000.
	uuids := Set new: count.
	count timesRepeat: [ uuids add: generator next ].
	self assert: uuids size equals: count! !

!NeoUUIDGeneratorTests methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:38'!
testUniquenessRandom
	| uuids count |
	count := 100000.
	uuids := Set new: count.
	count timesRepeat: [ uuids add: generator nextRandom ].
	self assert: uuids size equals: count! !

!NeoUUIDGeneratorTests methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/8/2016 18:45'!
timeFromUUID: uuid
	| bytes in |
	in := uuid readStream.
	bytes := ByteArray streamContents: [ :out |
		8 timesRepeat: [ out nextPut: in next ] ].
	^ bytes reverse asInteger ! !

Object subclass: #NeoUUIDGenerator
	instanceVariableNames: 'node random counter lock'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Neo-UUID'!

!NeoUUIDGenerator commentStamp: 'SvenVanCaekenberghe 12/11/2016 11:11' prior: 0!
I am NeoUUIDGenerator, I generate UUIDs.

An RFC4122 Universally Unique Identifier (UUID) is an opaque 128-bit number that can be used for identification purposes. Concretely, a UUID is a 16 element byte array.

The intent of UUIDs is to enable distributed systems to uniquely identify information without significant central coordination. In this context the word unique should be taken to mean "practically unique" rather than "guaranteed unique".
 
I generate UUIDs similar, in spirit, to those defined in RFC4122, though I use version 0 to indicate that I follow none of the defined versions, but a custom scheme instead. This does not matter much, if at all, in practice. This is the default.

Alternatively, I can generate pure random, fully conformant, version 4, UUIDs. These are simply 16 random bytes.

In my custom scheme, I try to conform to the following aspects:
 - each 'node' (machine, image, instance) should generate unique UUIDs
 - even when generating UUIDs at a very fast rate, they should remain unique
 - be fast and efficient

To achieve this goal, I
- take several aspects into account to generate a unique node ID
- combine a clock, a counter and some random bits
- hold a state, protected for multi user access

I can generate about 500K UUIDs per second.

Implementation:

Although a UUID should be seen as totally opaque, here is the concrete way I generate one:
- the first 8 bytes are the microsecond clock value with the smallest quantity first; this means that the later of these 8 bytes will be identical when generated with small(er) timespans; within the same microsecond, the full first 8 bytes will be identical
- the next 2 bytes represent a counter with safe overflow, held as protected state inside me; after 2*16 this value will repeat; the counter initalizes with a random value
- the next 2 bytes are simply random, based on the system PRNG, Random
- the final 4 bytes represent the node ID; the node ID is unique per instance of me, across OS environments where the image might run; the node ID is the MD5 hash of a string that is the concatenation of several elements (see #computeNodeIdentifier)
 
Some bits are set to some predefined value, to indicate the variant and version (see #set[Custom|Random]VariantAndVersion:)

Usage:

  NeoUUIDGenerator next.
  NeoUUIDGenerator default next.
  NeoUUIDGenerator new next.

Extra API:

  NeoUUIDGenerator default nextRandom.
  NeoUUIDGenerator default nextRandomByteArrayOfSize: 64.
  NeoUUIDGenerator default randomGenerator next.

Sharing an instance is more efficient and correct.
Instances should be reset (re-initialized or replaced) whenever the image comes up.

See also:

  http://en.wikipedia.org/wiki/UUID
  https://tools.ietf.org/html/rfc4122
!

!NeoUUIDGenerator class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/9/2016 10:51'!
default
	"Return the default UUID generator.
	Sharing an instance is more efficient and correct."
	
	^ Default ifNil: [ Default := self new ]! !

!NeoUUIDGenerator class methodsFor: 'initialization' stamp: 'SvenVanCaekenberghe 2/10/2016 21:00'!
initialize
	"Our default instance must be reset at startUp"
	
	self environment 
		at: #SessionManager 
		ifPresent: [ :manager | manager default registerNetworkClassNamed: self name ]
		ifAbsent: [ Smalltalk addToStartUpList: self ]! !

!NeoUUIDGenerator class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/9/2016 10:51'!
next
	"Return a UUID using the shared generator"
	
	"self next"
	
	"[ NeoUUIDGenerator next ] bench"
	
	^ self default next! !

!NeoUUIDGenerator class methodsFor: 'system startup' stamp: 'SvenVanCaekenberghe 2/9/2016 10:51'!
startUp
	"Whenever the image comes up, reset so that a new node ID is computed."
	
	Default := nil! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/8/2016 16:52'!
clockOn: stream
	| microseconds |
	microseconds := Time microsecondClockValue.
	1 to: 8 do: [ :each |
		stream nextPut: (microseconds digitAt: each) ]! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/11/2016 12:07'!
computeNodeIdentifier
	| nodeData |
	nodeData := String streamContents: [ :out |
		Smalltalk at: #NetNameResolver ifPresent: [ :netNameResolver |
			out print: netNameResolver localHostAddress.
			out print: netNameResolver localHostName].
		out print: SystemVersion current.
		out print: Smalltalk vm imagePath.
		out print: Smalltalk commandLine options.
		out print: Smalltalk session hash.
		out print: self hash ].
	^ (MD5 hashMessage: nodeData) hash asByteArrayOfSize: 4! !

!NeoUUIDGenerator methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 2/8/2016 20:04'!
initialize
	super initialize.
	random := Random new.
	lock := Semaphore forMutualExclusion.
	node := self computeNodeIdentifier.
	counter := self nextRandom16! !

!NeoUUIDGenerator methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:13'!
next
	"Return the next UUID, see my class comment"
	
	^ self nextCustom! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 8/31/2013 15:00'!
nextCounter16
	counter := (counter + 1) \\ 16rFFFF.
	^ counter! !

!NeoUUIDGenerator methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:16'!
nextCustom
	"Return the next UUID, following the custom scheme, see my class comment"
	
	| uuid |
	uuid := UUID nilUUID.
	self placeCustomFields: uuid.
	^ uuid! !

!NeoUUIDGenerator methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/10/2016 20:08'!
nextRandom
	"Return the next UUID, following the random scheme, see my class comment"

	| uuid |
	uuid := UUID nilUUID.
	self placeRandomFields: uuid.
	^ uuid! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2016 11:30'!
nextRandom16
	"Return a random integer in the interval [0,16rFFFF]"
	
	^ (random nextInt: 16r10000) - 1! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2016 11:30'!
nextRandom8
	"Return a random integer in the interval [0,16rFF]"

	^ (random nextInt: 16r100) - 1! !

!NeoUUIDGenerator methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/10/2016 19:18'!
nextRandomByteArrayOfSize: count
	"Return a ByteArray of count random bytes"
	
	| bytes |
	bytes := ByteArray new: count.
	1 to: count do: [ :each |
		bytes at: each put: self nextRandom8 ].
	^ bytes! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/10/2016 20:08'!
placeCustomFields: uuid
	"Fill the fields of uuid following my custom scheme, 
	effectively generating a new UUID in place, see my class comment"
	
	| byteStream |
	lock critical: [ 
		byteStream := uuid writeStream.
		self clockOn: byteStream.
		byteStream
			uint16: self nextCounter16;
			uint16: self nextRandom16;
			nextPutAll: node.
		self setCustomVariantAndVersion: uuid ] ! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/10/2016 20:09'!
placeRandomFields: uuid
	"Fill the fields of uuid following the random scheme, 
	effectively generating a new UUID in place, see my class comment"

	lock critical: [ 
		1 to: uuid size do: [ :each |
			uuid at: each put: self nextRandom8 ].
		self setRandomVariantAndVersion: uuid ]! !

!NeoUUIDGenerator methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 2/9/2016 10:46'!
randomGenerator
	"Return the Random generator that I use"
	
	^ random! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/10/2016 19:58'!
setCustomVariantAndVersion: uuid
	"Set the version (top four most significant bits) to '0000', an undefined version in RFC 4122"
	uuid at: 7 put: ((uuid at: 7) bitAnd: 2r1111).
	"Set the variant (top two most significant bits) to '10', the RFC 4122 variant"
	uuid at: 9 put: (((uuid at: 9) bitAnd: 2r111111) bitOr: 2r10000000)! !

!NeoUUIDGenerator methodsFor: 'private' stamp: 'SvenVanCaekenberghe 2/10/2016 19:57'!
setRandomVariantAndVersion: uuid
	"Set the version (top four most significant bits) to '0100', version 1 in RFC 4122"
	uuid at: 7 put: (((uuid at: 7) bitAnd: 2r1111) bitOr: 2r01000000).
	"Set the variant (top two most significant bits) to '10', the RFC 4122 variant"
	uuid at: 9 put: (((uuid at: 9) bitAnd: 2r111111) bitOr: 2r10000000)! !
NeoUUIDGenerator initialize!
